import copy
import json
import math
import os
import pdb
import pickle
import shutil
import sys
import time

import torch
import torch.nn.functional
import numpy as np
import torchvision
import torch.utils.data
import matplotlib.pyplot as plt
import wandb
from torch.hub import download_url_to_file
from tqdm import tqdm  # pip3 install tqdm
plt.rcParams["figure.figsize"] = (15, 15)
plt.style.use('dark_background')
import torch.utils.data
import scipy.misc
import scipy.ndimage
import sklearn.decomposition
import argparse # pip3 install argparse
import os
import torch.nn.functional as F
os.environ['WANDB_START_METHOD'] = 'thread'
os.environ['WANDB_API_KEY'] = ''
wandb.login()

parser = argparse.ArgumentParser(add_help=False)

parser.add_argument('-num_epochs', default=20, type=int)
parser.add_argument('-batch_size', default=64, type=int)
parser.add_argument('-learning_rate', default=1e-4, type=float)

parser.add_argument('-z_size', default=64, type=int)

parser.add_argument('-timesteps', default=100, type=int)
parser.add_argument('-noise_start', default=1e-5, type=float)
parser.add_argument('-noise_end', default=1e-1, type=float)

parser.add_argument('-sweep_id', default='', type=str)
parser.add_argument('-is_debug', default=True, type=lambda x: (str(x).lower() == 'true'))


args, _ = parser.parse_known_args()


BATCH_SIZE = args.batch_size
EPOCHS = args.num_epochs
TRAIN_TEST_SPLIT = 0.8

DEVICE = 'cpu'
if torch.cuda.is_available():
    DEVICE = 'cuda'

print("Using device:", DEVICE)

MAX_LEN = 200 # limit max number of samples otherwise too slow training (on GPU use all samples / for final training)
if DEVICE != 'cpu':
    MAX_LEN = None

# empty to include all
INCLUDE_CLASSES = [
    'Pomegranate',
    'Orange',
    'Avocado',
    'AppleGrannySmith',
    'Banana',
    'CucumberRipe'
]

class DatasetFruits(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__()
        dataset_name = 'fruits_32.pkl'
        path_dataset = f'../data/{dataset_name}'
        if not os.path.exists(path_dataset):
            os.makedirs('../data', exist_ok=True)
            download_url_to_file(
                f'http://share.yellowrobot.xyz/1630528570-intro-course-2021-q4/{dataset_name}',
                path_dataset,
                progress=True
            )
        with open(path_dataset, 'rb') as fp:
            X_tmp, Y_tmp, self.labels = pickle.load(fp)

        X = []
        Y = []
        if len(INCLUDE_CLASSES) == 0:
            X = X_tmp
            Y = Y_tmp
        else:
            for x, y in zip(X_tmp, Y_tmp):
                y_label = self.labels[y]
                if y_label in INCLUDE_CLASSES:
                    X.append(x)
                    Y.append(y)

        X = torch.from_numpy(np.array(X)).float()
        self.X = X.permute(0, 3, 1, 2)

        self.input_size = self.X.size(-1)
        Y = torch.LongTensor(Y)
        self.Y = Y.unsqueeze(dim=-1)

    def __len__(self):
        if MAX_LEN:
            return MAX_LEN
        return len(self.X)

    def __getitem__(self, idx):
        x = self.X[idx] / 255.0
        y_label = self.Y[idx]
        return x, y_label

dataset_full = DatasetFruits()
train_test_split = int(len(dataset_full) * TRAIN_TEST_SPLIT)
dataset_train, dataset_test = torch.utils.data.random_split(
    dataset_full,
    [train_test_split, len(dataset_full) - train_test_split],
    generator=torch.Generator().manual_seed(0)
)

data_loader_train = torch.utils.data.DataLoader(
    dataset=dataset_train,
    batch_size=BATCH_SIZE,
    shuffle=True,
    drop_last=(len(dataset_train) % BATCH_SIZE < 6)
)

data_loader_test = torch.utils.data.DataLoader(
    dataset=dataset_test,
    batch_size=BATCH_SIZE,
    shuffle=False,
    drop_last=(len(dataset_test) % BATCH_SIZE < 6)
)

sweep_config = {
    'method': 'bayes'
}
metric = {
    'name': 'test_loss',
    'goal': 'minimize'
    }
sweep_config['metric'] = metric
parameters_dict = {
    'Z_SIZE': {
        'values': [32, 64, 128]
    },
    'LEARNING_RATE': {
        'values': [1e-3, 1e-4, 1e-5]
    },
    'T_TIME_STEPS': {
        'values': [100, 200]
    },
    'NOISE_START': {
        'values': [1e-4, 1e-5, 1e-6]
    }
}

sweep_config['parameters'] = parameters_dict


def train():
    LEARNING_RATE = args.learning_rate
    Z_SIZE = args.z_size
    T_TIME_STEPS = args.timesteps
    NOISE_START = args.noise_start
    NOISE_END = args.noise_end

    run = None
    if not args.is_debug:
        run = wandb.init()
        LEARNING_RATE = wandb.config.LEARNING_RATE
        Z_SIZE = wandb.config.Z_SIZE
        T_TIME_STEPS = wandb.config.T_TIME_STEPS
        NOISE_START = wandb.config.NOISE_START
        NOISE_END = wandb.config.NOISE_END

    # calculate noise scheduler
    # TODO
    betas = 0
    alphas = 0
    alphas_cumprod = 0
    sqrt_alphas_cumprod = 0
    one_minus_alphas_cumprod = 0

    def scale_rgb_float_to_3sigma(x):
        result = 3.5 * (x - 0.5) * 2
        return result

    def scale_3sigma_to_rgb_float(x):
        is_squeeze = False
        if x.ndim == 3:
            is_squeeze = True
            x = x[None, ...]
        # min max scaling
        x_min = x.reshape(x.size(0), 3, -1).min(dim=-1)[0]
        x_max = x.reshape(x.size(0), 3, -1).max(dim=-1)[0]
        x_min = x_min[:, :, None, None]
        x_max = x_max[:, :, None, None]
        result = (x - x_min) / (x_max - x_min)

        if is_squeeze:
            result = result[0]
        return result

    def forward_diffusion(x_0, t):
        result = torch.randn_like(x_0)
        #TODO
        return result

    # i_step = int(T_TIME_STEPS / 10)
    # i = 0
    # for t in range(T_TIME_STEPS-1, 0, -i_step):
    #     i += 1
    #     plt.subplot(1, 10, i)
    #     x_img = dataset_full[0][0]
    #     x_img = forward_diffusion(x_img, t)
    #     plt.imshow(scale_3sigma_to_rgb_float(scale_rgb_float_to_3sigma(x_img)).permute(1, 2, 0))
    # plt.show()
    # exit()

    class PositionalEncoding(torch.nn.Module):
        def __init__(self, num_embeddings, embedding_dim):
            super().__init__()

            pe = torch.zeros(num_embeddings, embedding_dim).to(DEVICE)
            position = torch.arange(0, num_embeddings, dtype=torch.float).unsqueeze(1)
            div_term = torch.exp(torch.arange(0, embedding_dim, 2).float() * (-np.log(10000.0) / embedding_dim))
            pe[:, 0::2] = torch.sin(position * div_term)
            pe[:, 1::2] = torch.cos(position * div_term)
            self.pe = pe

        def forward(self, idxes):
            return self.pe[idxes, :]


    class ResBlock(torch.nn.Module):
        def __init__(self, in_channels, out_channels, stride=1, groups=None):
            super().__init__()
            self.conv1 = torch.nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False)
            self.gn1 = torch.nn.GroupNorm(num_channels=out_channels, num_groups=math.ceil(out_channels/2) if groups is None else groups)
            self.conv2 = torch.nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1, bias=False)
            self.gn2 = torch.nn.GroupNorm(num_channels=out_channels, num_groups=math.ceil(out_channels/2) if groups is None else groups)
            self.is_bottleneck = False
            if stride != 1 or in_channels != out_channels:
                self.is_bottleneck = True
                self.shortcut = torch.nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False)

        def forward(self, x):
            residual = x
            out = self.conv1(x)
            out = F.relu(out)
            out = self.gn1(out)
            out = self.conv2(out)
            if self.is_bottleneck:
                residual = self.shortcut(x)
            out += residual
            out = F.relu(out)
            out = self.gn2(out)
            return out

    class Model(torch.nn.Module):
        def __init__(self):
            super().__init__()

            self.pos_enc = PositionalEncoding(num_embeddings=T_TIME_STEPS, embedding_dim=Z_SIZE)

            self.maxpool = torch.nn.MaxPool2d(2)
            self.upsample = torch.nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)

            self.conv_down_1 = ResBlock(3, Z_SIZE)
            self.conv_down_2 = ResBlock(Z_SIZE, Z_SIZE)
            self.conv_down_3 = ResBlock(Z_SIZE, Z_SIZE)
            self.conv_down_4 = ResBlock(Z_SIZE, Z_SIZE)

            self.conv_middle = ResBlock(Z_SIZE, Z_SIZE)

            self.conv_up_4 = ResBlock(Z_SIZE, Z_SIZE)
            self.conv_up_3 = ResBlock(Z_SIZE, Z_SIZE)
            self.conv_up_2 = ResBlock(Z_SIZE, Z_SIZE)
            self.conv_up_1 = ResBlock(Z_SIZE, 3, groups=1)

        def forward(self, x, t):
            # TODO pos encoding

            skip_1 = self.conv_down_1.forward(x)
            out = self.maxpool.forward(skip_1)

            skip_2 = self.conv_down_2.forward(out)
            out = self.maxpool.forward(skip_2)

            skip_3 = self.conv_down_3.forward(out)
            out = self.maxpool.forward(skip_3)

            skip_4 = self.conv_down_4.forward(out)

            out = self.conv_middle.forward(skip_4)

            out = out + skip_4
            out = self.conv_up_4.forward(out)

            out = self.upsample.forward(out)
            out = out + skip_3
            out = self.conv_up_3.forward(out)

            out = self.upsample.forward(out)
            out = out + skip_2
            out = self.conv_up_2.forward(out)

            out = self.upsample.forward(out)
            out = out + skip_1
            out = self.conv_up_1.forward(out)
            out = torch.tanh(out)
            return out


    model = Model()
    model = model.to(DEVICE)
    optimizer = torch.optim.Adam(model.parameters(), lr=LEARNING_RATE)


    metrics = {}
    for stage in ['train', 'test']:
        for metric in [
            'loss'
        ]:
            metrics[f'{stage}_{metric}'] = []

    for epoch in range(1, EPOCHS+1):

        metrics_epoch = {key: [] for key in metrics.keys()}

        for data_loader in [data_loader_train, data_loader_test]:

            stage = 'train'
            if data_loader == data_loader_test:
                stage = 'test'
                torch.set_grad_enabled(False)
                model = model.eval()
            else:
                torch.set_grad_enabled(True)
                model = model.train()

            for x, y_label in tqdm(data_loader, desc=stage):

                x = scale_rgb_float_to_3sigma(x).to(DEVICE)
                t = torch.randint(0, T_TIME_STEPS, (x.size(0),)).to(DEVICE)
                x_noisy = forward_diffusion(x, t)

                #TODO
                loss = 0

                metrics_epoch[f'{stage}_loss'].append(loss.item())

                if data_loader == data_loader_train:
                    loss.backward()
                    optimizer.step()
                    optimizer.zero_grad()


        metrics_strs = []
        wandb_dict = {}
        for key in metrics_epoch.keys():
            if '_z' not in key and '_labels' not in key:
                value = np.mean(metrics_epoch[key])
                metrics[key].append(value)
                wandb_dict[key] = value
                metrics_strs.append(f'{key}: {round(value, 2)}')
        print(f'epoch: {epoch} {" ".join(metrics_strs)}')

        half_steps = int(T_TIME_STEPS * 0.0)
        x_noisy = forward_diffusion(x, T_TIME_STEPS - half_steps - 1)
        np_x = scale_3sigma_to_rgb_float(x_noisy).cpu().permute(0, 3, 2, 1).data.numpy()
        i_step = int((T_TIME_STEPS - half_steps) / 20)
        np_y_prim_steps = []
        for i in tqdm(range(half_steps, T_TIME_STEPS-1), desc='denoising'):
            t = torch.ones(x.size(0), dtype=torch.long).to(DEVICE) * i

            # TODO denoising inference

            np_y_prim = scale_3sigma_to_rgb_float(x_noisy).cpu().permute(0, 3, 2, 1).data.numpy()

            if i % i_step == 0:
                np_y_prim_steps.append(np_y_prim)
        np_y_prim_steps = np.stack(np_y_prim_steps)

        for i in range(np_x.shape[0]):
            x_progression = [np_x[i]]
            x_progression.extend(np_y_prim_steps[:, i])
            wandb_dict[f'progression_{i}'] = [wandb.Image(img) for img in x_progression]

        if not args.is_debug:
            wandb.log(wandb_dict)

    if not args.is_debug:
        run.finish()

if __name__ == "__main__":
    if not args.is_debug:
        project_name = "2023-06-12-diffusion-models-sweep"
        sweep_id = args.sweep_id
        if sweep_id == '':
            sweep_id = wandb.sweep(sweep_config, project=project_name)
        print(f"sweep_id: {sweep_id}") # can be used on another machine to join the sweep
        wandb.agent(sweep_id, function=train, project=project_name)
    else:
        train()